// Functions to Input and Outupt Serial data.
// =========================================

// Test_Serial_4() reads SeaTalk1 data and uses TestSeaTalk() defined in ST_NMEA ino file. It
// returns  line_4 = 1 if SeaTalk datagram was received and converted to Nmea0183 in nmea_4[]
// and, of course, if the nmea_4[] sentence passes filter 4.
// If line_4 was set, we send this nmea_4[] sentence to the Serial port through filter 3
// using function Output_4(). 

void Test_Serial_4() {     // Serial port #4
  if ( DebugMode ) { 
    TestSeaTalk_DB();
    return;
  }
  TestSeaTalk();
}

void Output_4() {
  if ( Skip_3 < FltSkip_3 ) {
    Skip_3 = Skip_3 + 1; 
  }
  else {
    Skip_3 = 0; 
    if ( CSum_3) {
      if ( !Is_CKS_Wrong(nmea_4) ) { if ( CheckNmeaFlt_3( 4 ) ) { Serial.print( nmea_4 ); } }
    }
    else { if ( CheckNmeaFlt_3( 4 ) ) { Serial.print( nmea_4 ); } }    
  }
  line_4 = 0;          // reset buffer & flag
  i_4 = 0;
}


// Test_Serial_1() reads Nmea sentences on the Serial interface. The sentences are filtered
// by filter 1 and placed in char array nmea_1[]. The function sets line_1 = 1 to signal
// that a sentence was received and is ready to be sent out by Output_(1). This
// function is defined in NMEA_ST.INO and passes the sentence through filter 5 and
// tries to convert and send it into the SeaTalk bus

void Test_Serial_1() {
  byte b;
  while (Serial.available() > 0) {    // get the character
    b = Serial.read();
    if ( Isnmea_1 ) {
      nmea_1[i_1] = b; 
      i_1 = i_1 + 1;
      if (i_1 == 6) { 
        if ( CheckNmeaFlt_1() ) { if ( Led1 == false ) { Led1 = true; } }
        else { Isnmea_1 = false; return; }
      }
      if (b == 0x0D) {        //0D is CR
        if ( i_1 > 8 ) {
          nmea_1[i_1] = 0x0A;
          nmea_1[i_1 + 1] = 0x00; 
          if ( CSum_1 ) {
            if ( Is_CKS_Wrong(nmea_1) ){ Isnmea_1 = false; return; }
          }   
          line_1 = 1;      
          Isnmea_1 = false;   
          return;
        }
        else { Isnmea_1 = false; return; } 
      }
    }
    if ( Isnmea_1 == false ) {
      if (( b == 0x21 ) || ( b == 0x24 )) {      //  test for ! or $  (AIS or NMEA)
        if ( Skip_1 < FltSkip_1 ) {
          Skip_1 = Skip_1 + 1; }
        else {
          Skip_1 = 0;
          nmea_1[0] = b; 
          i_1 = 1;
          Isnmea_1 = true;          
        }
      } 
    }
    if (i_1 > 80) { Isnmea_1 = false;  return; }   // line too long   
  }
}

char HexTable[128];

boolean Is_CKS_Wrong( char nmea[] ){
  // first get the checksum at the end of sentence
  int len = strlen(nmea) - 5;
  int H2 = HexTable[ nmea[len + 1] ];
  if ( H2 > 15 ) { return true; }
  int H1 = HexTable[ nmea[len + 2] ];
  if ( H1 > 15 ) { return true; }  
  H2 = 16 * H2 + H1; 
  // now evaluate the checksum 
  int XOR = 0;
  for (int i = 1; i < len; i++) {
    XOR = XOR ^ nmea[i];
  }
  if ( H2 == XOR ) { return false; }
  return true;
}

void BuildHexTable() {
  // this table is to convert hex value in ASCII format to an integer 
  // fill all the table with 16 - not valid value
  for ( int i = 0; i < 128; i++) {
    HexTable[i] = 16 ;   
  }
  // now fill the valid characters
  HexTable[48] = 0;   HexTable[49] = 1;   HexTable[50] = 2;
  HexTable[51] = 3;   HexTable[52] = 4;   HexTable[53] = 5;  
  HexTable[54] = 6;   HexTable[55] = 7;   HexTable[56] = 8;  
  HexTable[57] = 9;  
  // Capitals
  HexTable[65] = 10;  HexTable[66] = 11;  HexTable[67] = 12;  
  HexTable[68] = 13;  HexTable[69] = 14;  HexTable[70] = 15;
  // small
  HexTable[97] = 10;  HexTable[98] = 11;  HexTable[99] = 12;  
  HexTable[100] = 13; HexTable[101] = 14; HexTable[102] = 15;
}
