
char *pTOKEN;  // global pointer of type char, used to get the fields in a nmea sentence
               // when a new sentence is processed we advance this pointer 7 positions 
               // after the beginning of the sentence such it points to the 1st field.
               // used by function void FillToken(char * ptr)

void Output_1() {
  // first thing to do is to check if the nmea sentence is to be skipped
  if ( Skip_5 < FltSkip_5 ) { 
    Skip_5 = Skip_5 + 1;  line_1 = 0; i_1 = 0;  return; 
  } // sentence ignored 
  Skip_5 = 0;
  // then check if CheckSum needs validation
  if ( CSum_5) {
    if ( Is_CKS_Wrong(nmea_1) ) {
      line_1 = 0; i_1 = 0;  return; 
    } // sentence ignored
  } 
  // now we check if the sentence passes the filter
  byte b = CheckNmeaFlt_5( 1 );
  if ( b == 0 ) { line_1 = 0; i_1 = 0;  return; }   // sentence is to be rejected
  // now we check if the sentence is convertible and if buffer has space for it
  if ( Parse_NMEA ( b ) == false ) { line_1 = 0; i_1 = 0;  return; } // sentence is not convertible or no space
  if (Output_ST == false ) { Output_ST_begin();  }
  line_1 = 0;          // reset buffer & flag
  i_1 = 0;
}


boolean Parse_NMEA( byte b ){
  switch (b) {
  case 10: return ParseNmea10(); break;      // this is for DBT - Depth below transducer
  case 12: return ParseNmea12(); break;      // this is for DPT - Depth of Water
  case 35: return ParseNmea35(); break;      // this is for MTW - Mean Temperature of Water
  case 36: return ParseNmea36(); break;      // this is for MWV - Wind Speed and Angle
  case 18: return ParseNmea18(); break;      // this is for GLL - Geographic Position - Latitude/Longitude
  case 42: return ParseNmea42(); break;      // this is for RMC - Recommended Minimum Navigation Information  
  default: return false; break;
  }
}

boolean ParseNmea10() { // this is for DBT
  // DBT - Depth below transducer
  //        1   2 3   4 5   6 7
  //        |   | |   | |   | |
  // $--DBT,x.x,f,x.x,M,x.x,F*hh<CR><LF>
  // 1) Depth, feet
  // 2) f = feet
  // 3) Depth, meters
  // 4) M = meters
  // 5) Depth, Fathoms
  // 6) F = Fathoms
  // 7) Checksum
  //  Example: $SDDBT,7.8,f,2.4,M,1.3,F*0D
  // first thing to do is to check if there is space in the buffer
  int nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return false; }   // sentence is lost
  //  00  02  YZ  XX XX  Depth below transducer: XXXX/10 feet
  // we define the char arrays to receive the fields that exist in the nmea sentence
  char FFF[6]; char XXX[2]; char MMM[6];
  pTOKEN = nmea_1; pTOKEN = pTOKEN + 7;  // pTOKEN points to Depth, feet
  char *p;  // pointer to the receiving char arrays
  p = FFF; if ( !FillToken(p) ) { return false; };
  uint16_t ff; 
  if ( FFF[0] == 0 ) {    // feet field is empty, so check meters
    p = XXX; if ( !FillToken(p) ) { return false; };
    p = MMM; if ( !FillToken(p) ) { return false; };
    if ( MMM[0] == 0 ) { return false; }  // meters field also empty
    else { ff = 32.808 * atof(p); } 
  }
  else { ff = 10 * atof(p); }
  ST_buffer[inBuf][0] = 0x05;          // length
  ST_buffer[inBuf][1] = 0x00;          // command
  ST_buffer[inBuf][2] = 0x02;
  ST_buffer[inBuf][3] = 0x00;
  ST_buffer[inBuf][4] = lowByte(ff);
  ST_buffer[inBuf][5] = highByte(ff);   
  inBuf = nextBuf;
  return true; 
}

boolean ParseNmea12() { // this is for DPT
  // DPT - Depth of Water
  //         1   2   3   4
  //         |   |   |   |
  //  $--DPT,x.x,x.x,x.x*hh<CR><LF>
  // 1) Depth, meters
  // 2) Offset from transducer;
  //    positive means distance from transducer to water line,
  //    negative means distance from transducer to keel
  // 3) Maximum range scale in use (NMEA 3.0 and above)
  // 4) Checksum
  // Example: $INDPT,2.3,0.0*46
  // first thing to do is to check if there is space in the buffer
  int nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return false; }   // sentence is lost
  //  00  02  YZ  XX XX  Depth below transducer: XXXX/10 feet
  // we define the char array to receive depth in meters
  char MMM[6];
  pTOKEN = nmea_1; pTOKEN = pTOKEN + 7;  // pTOKEN points to depth in meters
  char *p = MMM; if ( !FillToken(p) ) { return false; };
  if ( MMM[0] == 0 ) { return false; }
  ST_buffer[inBuf][0] = 0x05;          // length
  ST_buffer[inBuf][1] = 0x00;          // command
  ST_buffer[inBuf][2] = 0x02;
  ST_buffer[inBuf][3] = 0x00;
  uint16_t mm = 32.808 * atof(p);
  ST_buffer[inBuf][4] = lowByte(mm);
  ST_buffer[inBuf][5] = highByte(mm); 
  inBuf = nextBuf;  
  return true;
}


boolean ParseNmea18() { // this is for GLL - generates ST 50 51 and 54
  /*
  GLL - Geographic Position - Latitude/Longitude
  This is one of the sentences commonly emitted by GPS units.
         1       2 3        4 5         6 7   8
         |       | |        | |         | |   |
  $--GLL,llll.ll,a,yyyyy.yy,a,hhmmss.ss,a,m,*hh<CR><LF>
  1) Latitude
  2) N or S (North or South)
  3) Longitude
  4) E or W (East or West)
  5) Time (UTC)
  6) Status A - Data Valid, V - Data Invalid
  7) FAA mode indicator (NMEA 2.3 and later)
  8) Checksum
  Example: $GNGLL,4404.14012,N,12118.85993,W,001037.00,A,A*67
  */
  // first thing to do is to check if there is space in the buffer
  int nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return false; }   // sentence is lost
  // we define the char arrays to receive the fields that exist in the nmea sentence
  char LAT[12]; char N_S[2]; char LON[13]; char E_W[2]; char GMT[10]; char VAL[2];
  pTOKEN = nmea_1; pTOKEN = pTOKEN + 7;  // pTOKEN points to latitude
  char *p;  // pointer to the receiving char arrays
  p = LAT; if ( !FillToken(p) ) { return false; }; if ( LAT[0] == 0 ) { return false; }
  p = N_S; if ( !FillToken(p) ) { return false; }; if ( N_S[0] == 0 ) { return false; }
  p = LON; if ( !FillToken(p) ) { return false; }; if ( LON[0] == 0 ) { return false; }
  p = E_W; if ( !FillToken(p) ) { return false; }; if ( E_W[0] == 0 ) { return false; }
  p = GMT; if ( !FillToken(p) ) { return false; };
  p = VAL; if ( !FillToken(p) ) { return false; };
  if (VAL[0] != 65) { return false; }    // 65 is ASCII A which validates the sentence
  // now start with Latitude
  // =======================
  p = LAT; 
  uint8_t XX = atoi(p) / 100;
  p = p + 2;
  uint16_t YYYY = 100 * atof(p);
  // check for N or S      
  ( N_S[0] == 0x4E ) ? YYYY = YYYY & 0x7FFF : YYYY = YYYY | 0x8000 ;  //  0x4E is ASCII N
  // now we can generate datagram 50
  ST_buffer[inBuf][0] = 0x05;          // length
  ST_buffer[inBuf][1] = 0x50;          // command
  ST_buffer[inBuf][2] = 0x02;
  ST_buffer[inBuf][3] = XX;    
  ST_buffer[inBuf][4] = lowByte(YYYY); 
  ST_buffer[inBuf][5] = highByte(YYYY);
  inBuf = nextBuf;    
  // now check if there is space in the buffer for the next datagram (longitude - 51)
  nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return true; }   // return true for previous datagram to be sent
  // now process longitude
  // =====================
  p = LON; 
  XX = atoi(p) / 100;
  p = p + 3;
  YYYY = 100 * atof(p);
  // check for E or W
  ( E_W[0] == 0x57 ) ? YYYY = YYYY & 0x7FFF : YYYY = YYYY | 0x8000 ;  //  0x57 is ASCII W
  // now we can generate datagram 51
  ST_buffer[inBuf][0] = 0x05;          // length
  ST_buffer[inBuf][1] = 0x51;          // command
  ST_buffer[inBuf][2] = 0x02;
  ST_buffer[inBuf][3] = XX;    
  ST_buffer[inBuf][4] = lowByte(YYYY); 
  ST_buffer[inBuf][5] = highByte(YYYY);
  inBuf = nextBuf;  
  // now check if there is space in the buffer for the next datagram (GMT - 54)
  nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return true; }   // return true for previous datagrams to be sent
  // now we process GMT
  // ==================
  if ( GMT[0] != 0 ) {
    // the field exists - so generate datagram 54
    p = GMT;
    char H[1]; strncpy(H, p, 2); uint8_t HH = atoi(H);
    p = p + 2; strncpy(H, p, 2);  uint8_t MM = atoi(H);
    p = p + 2; strncpy(H, p, 2); uint8_t SS = atoi(H);
   // now we can generate datagram 54
    ST_buffer[inBuf][0] = 0x04;         // length
    ST_buffer[inBuf][1] = 0x54;         // command
    ST_buffer[inBuf][4] = HH;           // hours go to last byte! easy!
    uint8_t i = SS & 0x0F; i = i << 4; 
    ST_buffer[inBuf][2] = i | 0x01;
    i = SS & 0x30; i = i >> 4;
    uint8_t j = MM; j = j << 2;
    ST_buffer[inBuf][3] = ( i | j );     // it is done!
    inBuf = nextBuf;
  }
  return true;
}

boolean ParseNmea35() { // this is for MTW
  //        1   2 3
  //        |   | |
  // $--MTW,x.x,C*hh<CR><LF>
  // 1) Degrees
  // 2) Unit of Measurement, Celcius
  // 3) Checksum
  // Example: $INMTW,17.9,C*1B
  // first thing to do is to check if there is space in the buffer
  int nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return false; }   // sentence is lost
  // we define the char array to receive the field that exists in the nmea sentence
  char DEG[8];
  pTOKEN = nmea_1; pTOKEN = pTOKEN + 7;  // pTOKEN points to Degrees
  char *p = DEG; if ( !FillToken(p) ) { return false; }; if ( DEG[0] == 0 ) { return false; }
  uint16_t mm = 10 * atof(p);
  mm = mm + 100;
  ST_buffer[inBuf][0] = 0x04;          // length
  ST_buffer[inBuf][1] = 0x27;          // command
  ST_buffer[inBuf][2] = 0x01;
  ST_buffer[inBuf][3] = lowByte(mm);
  ST_buffer[inBuf][4] = highByte(mm);   
  inBuf = nextBuf;    
  return true;
}

boolean ParseNmea36() { // this is for MWV
  //        1   2 3   4 5 6
  //        |   | |   | | |
  // $--MWV,x.x,a,x.x,a,a*hh<CR><LF>
  // 1) Wind Angle, 0 to 360 degrees
  // 2) Reference, R = Relative, T = True
  // 3) Wind Speed
  // 4) Wind Speed Units, K/M/N
  // 5) Status, A = Data Valid
  // 6) Checksum
  // Example: $WIMWV,200.0,R,11.0,N,A*11
  // first thing to do is to check if there is space in the buffer
  //  long start_time = micros();  
  int nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return false; }   // sentence is lost
  // we define the char arrays to receive the fields that exist in the nmea sentence
  char AWA[8]; char REF[2]; char AWS[10]; char KMN[2]; char VAL[2];
  pTOKEN = nmea_1; pTOKEN = pTOKEN + 7;  // pTOKEN points to AWA
  char *p;  // pointer to the receiving char arrays
  p = AWA; if ( !FillToken(p) ) { return false; }; if ( AWA[0] == 0 ) { return false; }
  p = REF; if ( !FillToken(p) ) { return false; };  
  p = AWS; if ( !FillToken(p) ) { return false; }; if ( AWS[0] == 0 ) { return false; }
  p = KMN; if ( !FillToken(p) ) { return false; };
  p = VAL; if ( !FillTokenEnd(p) ) { return false; };              // there no "," after Status so we need to use "*"
  if (VAL[0] != 65) { return false; }    // 65 is ASCII A which validates the sentence
  // now start with AWA
  // ==================
  // 10  01  XX  YY  Apparent Wind Angle: XXYY/2 degrees right of bow
  p = AWA; 
  uint16_t awa = 2 * atof(p);
  ST_buffer[inBuf][0] = 0x04;          // length
  ST_buffer[inBuf][1] = 0x10;          // command
  ST_buffer[inBuf][2] = 0x01;
  ST_buffer[inBuf][4] = lowByte(awa);
  ST_buffer[inBuf][3] = highByte(awa);
  inBuf = nextBuf;    
  // now check if there is space in the buffer for the next datagram (wind speed - 11 )
  nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return true; }   // return true for previous datagram to be sent
  // now process AWS
  // ===============
  // 11  01  XX  0Y  Apparent Wind Speed: (XX & 0x7F) + Y/10 Knots
  p = AWS;
  float aws = atof(p);
  if ( KMN[0] == 75 ) { aws = 0.53996 * aws;  } // "K" is ASCII 75   km/hour
  if ( KMN[0] == 77 ) { aws = 1.94384 * aws;  } // "M" is ASCII 77    m/sec
  uint8_t i = aws;
  uint8_t k = 10 * aws;
  k = k - ( 10 * i );
  ST_buffer[inBuf][0] = 0x04;          // length
  ST_buffer[inBuf][1] = 0x11;          // command
  ST_buffer[inBuf][2] = 0x01;
  ST_buffer[inBuf][3] = i;
  ST_buffer[inBuf][4] = k; 
  inBuf = nextBuf;  
  // long end_time = micros(); Serial.print( "Time in uS = "); Serial.println( end_time - start_time );
  return true;
}

boolean ParseNmea42() { // this is for RMC
  /*
  RMC - Recommended Minimum Navigation Information
  This is one of the sentences commonly emitted by GPS units.
                                                            12
         1         2 3       4 5        6  7   8   9    10 11|  13
         |         | |       | |        |  |   |   |    |  | |   |
  $--RMC,hhmmss.ss,A,llll.ll,a,yyyyy.yy,a,x.x,x.x,xxxx,x.x,a,m,*hh<CR><LF>
   1) Time (UTC)
   2) Status, V = Navigation receiver warning
   3) Latitude
   4) N or S
   5) Longitude
   6) E or W
   7) Speed over ground, knots
   8) Track made good, degrees true
   9) Date, ddmmyy
  10) Magnetic Variation, degrees
  11) E or W
  12) FAA mode indicator (NMEA 2.3 and later)
  13) Checksum
  Example: $GPRMC,120004.03,A,3905.84900,N,02633.42800,E,0000.0,330.0,011019,0.0,W,A,S*64
  */  
  // first thing to do is to check if there is space in the buffer
  int nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return false; }   // sentence is lost
  // we define the char arrays to receive the fields that exist in the nmea sentence
  char GMT[10]; char VAL[2]; char LAT[12]; char N_S[2]; 
  char LON[13]; char E_W[2]; char SOG[6]; char COG[6]; char DAT[7];
  pTOKEN = nmea_1; pTOKEN = pTOKEN + 7;  // pTOKEN points to GMT time
  char *p;  // pointer to the receiving char arrays
  p = GMT; if ( !FillToken(p) ) { return false; };
  p = VAL; if ( !FillToken(p) ) { return false; };
  if (VAL[0] != 65) { return false; }    // 65 is ASCII A which validates the sentence
  p = LAT; if ( !FillToken(p) ) { return false; }; if ( LAT[0] == 0 ) { return false; }
  p = N_S; if ( !FillToken(p) ) { return false; }; if ( N_S[0] == 0 ) { return false; }
  p = LON; if ( !FillToken(p) ) { return false; }; if ( LON[0] == 0 ) { return false; }
  p = E_W; if ( !FillToken(p) ) { return false; }; if ( E_W[0] == 0 ) { return false; }
  p = SOG; if ( !FillToken(p) ) { return false; }; 
  p = COG; if ( !FillToken(p) ) { return false; }; 
  p = DAT; if ( !FillToken(p) ) { return false; }; 
  // now start with GMT
  // ==================
  if ( GMT[0] != 0 ) {
    // the field exists - so generate datagram 54
    p = GMT;
    char H[1]; strncpy(H, p, 2); uint8_t HH = atoi(H);
    p = p + 2; strncpy(H, p, 2);  uint8_t MM = atoi(H);
    p = p + 2; strncpy(H, p, 2); uint8_t SS = atoi(H);
   // now we can generate datagram 54
    ST_buffer[inBuf][0] = 0x04;         // length
    ST_buffer[inBuf][1] = 0x54;         // command
    ST_buffer[inBuf][4] = HH;           // hours go to last byte! easy!
    uint8_t i = SS & 0x0F; i = i << 4; 
    ST_buffer[inBuf][2] = i | 0x01;
    i = SS & 0x30; i = i >> 4;
    uint8_t j = MM; j = j << 2;
    ST_buffer[inBuf][3] = ( i | j );     // it is done!
    inBuf = nextBuf;
    // now check if there is space in the buffer for the next datagram (latitude - 51)
    nextBuf = inBuf + 1; 
    if (nextBuf == 8) { nextBuf = 0; }
    if ( nextBuf == outBuf ) { return true; }   // return true for previous datagram to be sent 
  }
  // now process latitude
  // ====================
  p = LAT; 
  uint8_t XX = atoi(p) / 100;
  p = p + 2;
  uint16_t YYYY = 100 * atof(p);
  // check for N or S      
  ( N_S[0] == 0x4E ) ? YYYY = YYYY & 0x7FFF : YYYY = YYYY | 0x8000 ;  //  0x4E is ASCII N
  // now we can generate datagram 50
  ST_buffer[inBuf][0] = 0x05;          // length
  ST_buffer[inBuf][1] = 0x50;          // command
  ST_buffer[inBuf][2] = 0x02;
  ST_buffer[inBuf][3] = XX;    
  ST_buffer[inBuf][4] = lowByte(YYYY); 
  ST_buffer[inBuf][5] = highByte(YYYY);
  inBuf = nextBuf;    
  // now check if there is space in the buffer for the next datagram (longitude - 51)
  nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return true; }   // return true for previous datagram to be sent
  // now process longitude
  // =====================
  p = LON; 
  XX = atoi(p) / 100;
  p = p + 3;
  YYYY = 100 * atof(p);
  // check for E or W
  ( E_W[0] == 0x57 ) ? YYYY = YYYY & 0x7FFF : YYYY = YYYY | 0x8000 ;  //  0x57 is ASCII W
  // now we can generate datagram 51
  ST_buffer[inBuf][0] = 0x05;          // length
  ST_buffer[inBuf][1] = 0x51;          // command
  ST_buffer[inBuf][2] = 0x02;
  ST_buffer[inBuf][3] = XX;    
  ST_buffer[inBuf][4] = lowByte(YYYY); 
  ST_buffer[inBuf][5] = highByte(YYYY);
  inBuf = nextBuf;  
  // now check if there is space in the buffer for the next datagram (SOG - 52)
  nextBuf = inBuf + 1; 
  if (nextBuf == 8) { nextBuf = 0; }
  if ( nextBuf == outBuf ) { return true; }   // return true for previous datagrams to be sent
  // now process SOG
  // ==============
  if ( SOG[0] != 0 ) {
    // the field exists - so generate datagram 52
    p = SOG;
    //  52  01  XX  XX  Speed over Ground: XXXX/10 Knots
    uint16_t sog = 10 * atof(p);
    ST_buffer[inBuf][0] = 0x04;          // length
    ST_buffer[inBuf][1] = 0x52;          // command
    ST_buffer[inBuf][2] = 0x01;   
    ST_buffer[inBuf][3] = lowByte(sog); 
    ST_buffer[inBuf][4] = highByte(sog);
    inBuf = nextBuf;  
    // now check if there is space in the buffer for the next datagram (COG - 53)
    nextBuf = inBuf + 1; 
    if (nextBuf == 8) { nextBuf = 0; }
    if ( nextBuf == outBuf ) { return true; }   // return true for previous datagrams to be sent
  }
  // now process COG
  // ==============
  if ( COG[0] != 0 ) {
    // the field exists - so generate datagram 53
    p = COG;
    // 53  U0  VW      Course over Ground (COG) in degrees
    uint16_t cog = atoi(p);
    ST_buffer[inBuf][0] = 0x03;          // length
    ST_buffer[inBuf][1] = 0x53;          // command
    uint16_t ii = cog / 90; 
    uint16_t kk = ii;
    uint16_t jj = cog % 2;
    if (jj) { kk = (kk | B00001000); }
    ST_buffer[inBuf][2] = kk << 4;
    jj = cog - ii * 90;
    ST_buffer[inBuf][3] = jj / 2; 
    inBuf = nextBuf;  
    // now check if there is space in the buffer for the next datagram (Date - 56)
    nextBuf = inBuf + 1; 
    if (nextBuf == 8) { nextBuf = 0; }
    if ( nextBuf == outBuf ) { return true; }   // return true for previous datagrams to be sent
  }
  // now process DATE
  // ================
  if ( DAT[0] != 0 ) {
    // the field exists - so generate datagram 56
    p = DAT;
    // 56  M1  DD  YY  Date: YY year, M month, DD day in month
    char H[1]; strncpy(H, p, 2); uint8_t DD = atoi(H);
    p = p + 2; strncpy(H, p, 2); uint8_t MM = atoi(H);
    p = p + 2; strncpy(H, p, 2); uint8_t YY = atoi(H);
    ST_buffer[inBuf][0] = 0x04;          // length
    ST_buffer[inBuf][1] = 0x56;          // command
    ST_buffer[inBuf][2] = ( ( MM << 4 ) | 0x01 );
    ST_buffer[inBuf][3] = DD;
    ST_buffer[inBuf][4] = YY;
    inBuf = nextBuf;  
  }
  return true;
}

boolean FillToken(char * ptr) {
  /* the function receives a pointer to the char array where the field is to be stored.
   * It uses the global pointer pTOKEN that points to the starting of the field that we
   * want to extract. So we search for the next field delimiter "," which will be pointed
   * by p2 and we get the lenght of the field in the variable len. We copy len characters 
   * to the destination char array and terminate the array with a 0 (ZERO). So if a field 
   * is empty, len = 0 and ptr[0]=0. This can be used, later, to test if the field was empty.
   * Finally as p2 is pointing to a "," we increment it by 1 and copy it to pTOKEN so that
   * pTOKEN will be pointing to the starting of the next field.
   */
  char *p2 = strchr(pTOKEN, ',');
  if ( p2 == NULL ) { return false; }
  int len = p2 - pTOKEN;
  memcpy(ptr, pTOKEN, len);
  ptr[len]=0;
  pTOKEN = p2 + 1;
  return true;
}

boolean FillTokenEnd(char * ptr) {
  // same as before but searches for '*'
  char *p2 = strchr(pTOKEN, '*');
  if ( p2 == NULL ) { return false; }  
  int len = p2 - pTOKEN;
  memcpy(ptr, pTOKEN, len);
  ptr[len]=0;
  pTOKEN = p2 + 1;
  return true;
}
