// This is the code receiving Uart based on timer_1 in SeaTalk mode
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

int rxbuffSize;
int rxPin;
volatile byte rxBitCounter;
volatile uint8_t rxByte;
unsigned int inPos;
unsigned int outPos;
uint8_t *rxBuffer;             // declares a pointer to a uint8_t
hw_timer_t * timer_1 = NULL;
void IRAM_ATTR onRx();
void IRAM_ATTR onRxTimer();
// these for the 9th bit
uint8_t rxSign;
uint8_t *rxBufferSign;

void Serial_ST_setup() {
  // for timer1 used for input  
  rxPin = RX_ST_PIN;
  pinMode(rxPin, INPUT_PULLUP);  
  rxBuffer = (uint8_t*)malloc(RX_ST_SIZE);
  rxbuffSize = RX_ST_SIZE;
  rxBufferSign = (uint8_t*)malloc(RX_ST_SIZE);
}

void Serial_ST_begin() {
  inPos = outPos = 0; 
  timer_1 = timerBegin(1, 8, true);      // prescaler = 8      
  attachInterrupt(rxPin, onRx, FALLING);
}

void IRAM_ATTR onRx() { 
  detachInterrupt(rxPin);
  timerAttachInterrupt(timer_1, &onRxTimer, true);
  timerAlarmWrite(timer_1, init_ticks, true);
  timerAlarmEnable(timer_1);
  rxBitCounter = 0;  
  rxByte = 0;
  last_ST_time = micros();
}

void IRAM_ATTR onRxTimer() { 
  // ignore first shot and change alarm time on the 2nd shot
  if ( rxBitCounter <= 1 ) {
    if (rxBitCounter == 0) {
       rxBitCounter = 1;
       return;
    }  
    timerAlarmWrite(timer_1, bit_ticks, true);
  }  
  if (rxBitCounter == 9) {
    timerAlarmDisable(timer_1);
    // set 9th bit in byte rxSign
    if (digitalRead(rxPin)) { rxSign = 0xFF; }
    else { rxSign = 0x00; } 
    // and put rxByte received byte in the buffer
    int next = (inPos+1) % rxbuffSize;
    if (next != outPos) {
      rxBuffer[inPos] = rxByte;
      rxBufferSign[inPos] = rxSign;
      inPos = next;
    }
    attachInterrupt(rxPin, onRx, FALLING); 
    return;
  }
  //"put" the bit into rxByte
  rxByte >>= 1;
  if (digitalRead(rxPin)) { rxByte |= 0x80; }
  rxBitCounter = rxBitCounter + 1;
}

int Serial_ST_available() {
  int avail = inPos - outPos;
  if (avail < 0) avail += rxbuffSize;
  return avail;
}

uint16_t Serial_ST_read() {
   if ( inPos == outPos) return -1;
   uint16_t ch = rxBuffer[outPos];
   if ( rxBufferSign[outPos] ) { ch = ch | 256; }
   outPos = (outPos + 1) % rxbuffSize;
   return ch;
}

void Serial_ST_end() {
  detachInterrupt(rxPin);  
  if (timer_1 == NULL ) {return; }
  timerEnd(timer_1);
  timer_1 = NULL;  
}
