// This is the code to transmit ST data based on timer_0
// +++++++++++++++++++++++++++++++++++++++++++++++++++++

volatile int txPin;                // variable that holds the pin number used for transmission (is that necessary?)
volatile uint8_t txByteCounter;    // indicates the byte of the datagram that is being sent out
volatile uint8_t txBitCounter;     // indicates the  bit of the datagram that is being sent out
volatile boolean Bit9;             // if true the 9th bit should be 1 (as it is a command byte)
volatile boolean txLast;           // when we set txPin to send a bit we also set txLast with the same value. 
                                   // true is the same as HIGH and false is the same as LOW
                                   // For example, to send a bit value of 1 (data line=12V) txPin should be
                                   // LOW (set to false) so that the LED diode becomes OFF and the optotransistor is cut
                                   // bringing the data line voltage to 12V. So txLast is set to FALSE.
                                   // When time comes to send the next bit we read rxPin. As we had sent a bit equal to 1
                                   // and txLat was set as FaLSE we should see the data line with the same value of 12V
                                   // which menas that the receiving opto LED is OFF (bot cathode and anode at 12V)
                                   // so that the receiving optotransistor is cut-off and rxPin should be HIGH
                                   // or TRUE if no Bus Collision happened. THEREFORE we need to confirm if rxPin is
                                   // different from txLast. If they are equal someone changed the bus!                                   
//volatile byte ST_buffer[8][19];  // the 8 block buffer                              
//volatile int inBuf = 0;          // recycling at 8. So they vary as: 0 1 2 3 4 5 6 7 0 1 2 ....
//volatile int outBuf = 0;         // recycling at 8. So they vary as: 0 1 2 3 4 5 6 7 0 1 2 ....                      
//volatile uint8_t txByte;         // we copy here the byte (read from the buffer) that needs to be transmitted
//volatile uint8_t txSize;         // and here we copy its length
hw_timer_t * timer_0 = NULL;       // timer_0 is a pointer to a variable of type hw_timer_t
void IRAM_ATTR onTimer_0();        // this obliges the onTimer_0() to be placed on RAM by the compiler
void IRAM_ATTR  Output_ST_end();   // this obliges the onTimer_0() to be placed on RAM by the compiler

volatile long lasttime;
volatile long newtime;

void Output_ST_setup() {
  // for timer0 used for output 
  txPin = TX_ST_PIN; 
  pinMode(txPin, OUTPUT); 
  digitalWrite(txPin, LOW);         // iddle state is low (inverted by the transistor)
  inBuf = outBuf = 0; 
  Output_ST = false;          
}
                  
void IRAM_ATTR Output_ST_begin() {
  // wait until bus is idle for at least ST_wait uS
  while ( ( micros() - last_ST_time ) < ST_wait ) {  }
  if( LoopBack == false ) { Serial_ST_end(); }
  timer_0 = timerBegin(0, 8, true);
  timerAttachInterrupt(timer_0, &onTimer_0, true);
  timerAlarmWrite(timer_0, bit_ticks, true);
  timerAlarmEnable(timer_0);
  txBitCounter = 0;
  txByteCounter = 0;
  Bit9 = true;            // this will be used in the 1st byte of the datagram to set txPin to LOW or data line to 12V
                          // or Sign Bit to 1, so that the 9th bit command byte is 1. When Bit9 = false a Sign bit of 0 is sent!
  Output_ST = true;       // signal that bus is ON for output               
  // from this moment on timer0 will trigger every 208uS
}

void IRAM_ATTR onTimer_0() { 
  if ( inBuf == outBuf ) { return; }
  
  // +++++++ CASE BitCounter = 0 ++++++++
  if ( txBitCounter < 1 ) {                 // if we are starting with a new byte
    if (txByteCounter == 0) {               // check if it is the start of a new datagram
       txSize = ST_buffer[outBuf][0];       // if so get the size of the new datagram
       if (txSize == 0 ) { return; }
    }
    if ( txByteCounter == txSize ) {        // check if complete datagram has been sent.
      txByteCounter = 0;                    // prepare for next datagram 
      txBitCounter = 0;
      Bit9 = true;
      
      outBuf = outBuf + 1; 
      if (outBuf == 8) { outBuf = 0; }
      
      if ( inBuf == outBuf ) {              // check if there are no more datagrams to follow   
        Output_ST_end();                    // stop timer_0            
        Serial_ST_begin();                  // allow reception again
      }
      return;
    }
    // get the byte 
    txByteCounter = txByteCounter + 1;
    txByte = ST_buffer[outBuf][txByteCounter];
    // we test bus collision on txBitConter=0 here! 
    // the previous bus state was a stop bit (12V) and so the rxPin should be HIGH 
    // we need to test if a single return is enough!
    if ( digitalRead(rxPin) == LOW ) { txByteCounter = 0; return; }                                                         
    // now we need to send the START BIT (0V). So we must set txPin = HIGH or 1!
    txLast = HIGH ;          
    digitalWrite( txPin , txLast );
    txBitCounter = 1;
    return;
  }

  // we test bus collision on txBitConter=1 2 3 ... 10 here! 
  // we need to check if a single return is enough!
  if ( digitalRead(rxPin) == txLast ) { txBitCounter = txByteCounter = 0; return; }  // Bus Collision!

  // +++++++ CASE BitCounter = 1 ... 8  ++++++++  
  if ( txBitCounter < 9 ) { 
    txLast = HIGH ; 
    if( bitRead( txByte , txBitCounter - 1 ) ) { txLast = LOW ; }
    digitalWrite( txPin , txLast );
    txBitCounter = txBitCounter + 1 ;
    return;
  }

  // +++++++ CASE BitCounter 9 ++++++++
  if ( txBitCounter == 9 ) {      // now the SIGN bit
    txLast = HIGH ; 
    if ( Bit9 == true ) { txLast = LOW ; Bit9 = 0; }  // command Byte >>> SignBit=1 OR Data=12V or txPin=low
    digitalWrite( txPin , txLast );
    txBitCounter = 10 ;
    return;
  }
  // +++++++ CASE BitCounter 10  ++++++++  
  // always falls here. We send a StopBit=1 or Data=12V or txPin=LOW
  txLast = LOW;
  digitalWrite( txPin , txLast );
  txBitCounter = 0;
  return;
}


void IRAM_ATTR Output_ST_end() { 
  Output_ST = false;
  if ( timer_0 == NULL ) { return; }
  timerEnd(timer_0);
  timer_0 = NULL;  
}
